/* DYNAMIC SC DENSE CODE. Ranges approach version -- 
A dynamic word-based byte oriented compressor for text files based on 
(s,c)-Dense Code.

Programmed by Antonio Faria
Copyright (C) 2006  Nieves R. Brisaboa, Gonzalo Navarro, Antonio Faria and Jos R. Param
Author's contact: antonio.fari@gmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "decompressor.h"
#include "version.h"

/*------------------------------------------------------------------
 reads the vocabulary size and the size of the vector "top" from 
 the aulixiar file (if it exists)
 ---------------------------------------------------------------- */
void readSizeVocabularyandTop(char *filename,  long *vocSize,  long *topSize) {
	char fileNametmp[24];
	int file;
	unsigned long value;
	
	strcpy (fileNametmp,filename);
	strcat(fileNametmp,".voc");	
	
	if ( (file = open (fileNametmp, O_RDONLY)) ==-1) {
		fprintf(stderr,"\nError while opening file: %s\n\tUsing default vocabulary size instead\n",fileNametmp);
		return ;
	}
	read(file,  vocSize,  4);
	read(file,  topSize,  4);
	
	close (file);
//	*vocSize = 885630;
//	*topSize= 9485287;
}



/*------------------------------------------------------------------
 Initializes data structures
 ---------------------------------------------------------------- */
void initialize (unsigned long sizeVoc, unsigned long tamTop){
	unsigned long i;

	_memMgr = createMemoryManager(); //_memMgr.initialize();

	N=sizeVoc;
	words  = (t_word *) malloc(N * sizeof(t_word));
	zeroNode = 0;

	topBlock = (unsigned long *) malloc(tamTop * sizeof(unsigned long));

	//initializes top and last values for each block
	MAXFREQ = 1;
	topBlock[0] =zeroNode;

	S_value = 256;
 	C_value = B_value - S_value;

	T[0] = 0 ;
	T[1] = 0 ;
	T[2] = 0 ;

	I[0][0]=0;

}


/*------------------------------------------------------------------
 * oBytes: Returns the number of bytes needed to encode the word
 * ranked "i" when using s_i value as "S"
------------------------------------------------------------------ */
unsigned long oBytes (unsigned long s_i, unsigned long i_val) {
		unsigned long k, c_i;
		unsigned long last,pot;

		//fprintf(stdFile,"\nENTRO EN oBytes con s = %lu",s_i);
		//counts the number of bytes of the codeword to be generated
		k  	= 1;
		last  = pot = s_i;
		c_i	= B_value - s_i;

		while (last <= i_val) {
			pot *= c_i;
			last += pot;
			k++;
		}
		////fprintf(stdFile,"\n[ZERONODE=%lu]Size of i_val [%lu] when s= [%lu] is [%lu] \n",zeroNode,i_val,s_i,k);
		return k;
	}



/*------------------------------------------------------------------
 * takeAccount: Adds to TCS_{s-1}, TCS_{s}, TCS_{s+1} , the number of
 * bytes needed to encode the word ranked "i", depending on the "s" used
------------------------------------------------------------------ */
 void takeAccount () {

register int kkk;

//kkk=oBytes(S_value,j)-1;
kkk=K-1;

	 if (j>=I[kkk][1])  T[v[1][kkk]]++;
         else {
           if (j<I[kkk][0]) {

			T[v[0][kkk]]--;

            if (T[v[0][kkk]]<0) {
//				printf ("CAMBIO DE S: VALOR CAMBIA DE [%lu] A [%lu]\n",S_value,S_value+v[0][kkk]-1);
//				printf ("zeroNode vale [%lu]",zeroNode);
				S_value=S_value+v[0][kkk]-1;


//				if (S_value >=B_value) { S_value=B_value -1;}


				C_value=B_value-S_value;

//                                T[2-v[0][kkk]]=T[1];
//                                T[1]=T[v[0][kkk]];

				T[2-v[0][kkk]]=1;
				T[v[0][kkk]]=0;

				I[0][1]=W[S_value-1][0];
				I[1][0]=W[S_value+1][0];

				I[1][1]=W[S_value+1][1];
				I[2][0]=W[S_value-1][1];

				I[2][1]=W[S_value+1][2];
				I[3][0]=W[S_value-1][2];

				I[3][1]=W[S_value+1][3];
                        }
                }
         }
}




/*------------------------------------------------------------------
 * Performs the full decoding process over the compressed file.
------------------------------------------------------------------ */
void mainDecompressor (char* infile, char* outfile, int needToReadFile, unsigned long N_val, unsigned long N_val2) {
  
  unsigned char *palabra;
  unsigned long codeSize1;
  unsigned char *tmpW;
  register unsigned char abyte;
  BufferRead B;

	int f;
	/**/unsigned char *text,*top, *nextbuf;
    /**/unsigned char *pbeg,*pend,*codeStart;

	if (needToReadFile)
		readSizeVocabularyandTop(infile,(long *)&N_val,(long *)&N_val2);
	printf("\n VOC SIZE [%ld],     TOP SIZE [%ld] \n",N_val,N_val2);


// Inicializacin del buffer de lectura
	f = open (infile, O_RDONLY);
	if (f == -1)
		printf("Cannot read file %s",*infile);
	B = bufCreate(MAX_SIZE_OF_WORD+5);    //nmero de "bytes" que hay que dejar al final del buffer siempre, por culpa
													  //de que pudiese venir una nueva palabra"que fuese cortada".
	bufSetFile (B,f);

// Inicializacin del buffer de escritura
	outWriter = createWriter (outfile);
	
	initialize(N_val,N_val2); //initializes decompressor for up to N_value words
	
	fprintf(stderr, "\n\n\nSTARTING DECOMPRESSION...");


/************************************************************************/
/************************************************************************/


	if (bufEmpty(B) ) { return;}


		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
	    //fprintf(stderr,"\n top = [%ld]",((unsigned long) top) - ((unsigned long)B->data));


		while (( pbeg < pend )&& (zeroNode < S_value) )   {
				codeStart = pbeg;

				K=1;
				if ( (*pbeg) >= S_value ) {
					j = (*pbeg) - S_value;
					pbeg++;
					K++;

					while ((*pbeg) >= S_value) {

						j = j * C_value + ((*pbeg) - S_value);
						pbeg++;
						K++;
					}

					j = j * S_value + (*pbeg);
					if (K!=0)
					j += W[S_value][K-2];

				}
				else {
					j = (*pbeg);
				}


//				j=0;
//				while (*pbeg >=128) {
//					j = j * 128 + (*pbeg - 128);
//					pbeg++;
//				}
//				j = j * 128 + *pbeg;
//				j += TABLABASE[pbeg - codeStart];





				//fprintf(stderr,"\n decodificado=[%ld], bytes [%ld], con zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

				pbeg++;


	         if (j<zeroNode) {
					//	writes a word
					writeWord(outWriter,words[j].word, words[j].size);
				}
			 else { // j== zeronode

					codeStart = pbeg;
					while (*pbeg !='\0') {
						pbeg++;
					}

					//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

					{  register unsigned long len;
					   unsigned char *w;

	 				   len = pbeg-codeStart;

					   ////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
					   writeWord(outWriter,codeStart, len);

					   getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);

						w=words[j].word;
						register int i;
						for (i=0;i<len;i++) {  //copia tambien el '\0'
							*(w++)=*(codeStart++);
						}
						*w='\0';
						words[j].weight = 0;
						words[j].size=len;
					}

					pbeg++;
					zeroNode++;
	             // fprintf(stderr,"\n Chega ao final");

				}/*else == zeronode*/



		/**/	{    // HACE EL UPDATE
				register unsigned long f,p,i,h;
				register unsigned long temp;
				unsigned char *tmpW;

					p=j;
					f = words[p].weight;
					words[p].weight +=1;

					i=j;
					j = topBlock[f];

					temp = words[i].weight;
					words[i].weight = words[j].weight;
					words[j].weight = temp;

					tmpW = words[i].word;
					words[i].word = words[j].word;
					words[j].word = tmpW;

					temp = words[i].size;
					words[i].size = words[j].size;
					words[j].size = temp;

					topBlock[f]=j+1;

					if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
						topBlock[f+1] = 0;
						MAXFREQ++;
					}
					j=i;
				}
		/**/



			}/* while interno*/

		S_value --;
 	  	C_value  ++;
//		printf ("\n CAMBIO DE S: VALOR BAJA DE [%lu] A [%lu]\n",S_value+1,S_value);
//		printf ("zeroNode vale [%lu]",zeroNode);


/************************************************************************/
/* main 2 ... */
/************************************************************************/
/************************************************************************/


		while (( pbeg < pend ) && ( S_value > 210))  {
					codeStart = pbeg;

				K=1;
				if ( (*pbeg) >= S_value ) {
					j = (*pbeg) - S_value;
					pbeg++;
					K++;

					while ((*pbeg) >= S_value) {

						j = j * C_value + ((*pbeg) - S_value);
						pbeg++;
						K++;
					}

					j = j * S_value + (*pbeg);
					if (K!=0)
					j += W[S_value][K-2];

				}
				else {
					j = (*pbeg);
				}


//				j=0;
//				while (*pbeg >=128) {
//					j = j * 128 + (*pbeg - 128);
//					pbeg++;
//				}
//				j = j * 128 + *pbeg;
//				j += TABLABASE[pbeg - codeStart];





				//fprintf(stderr,"\n decodificado=[%ld], bytes [%ld], con zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

				pbeg++;


	         if (j<zeroNode) {
					//	writes a word
					writeWord(outWriter,words[j].word, words[j].size);
				}
			 else { // j== zeronode

					codeStart = pbeg;
					while (*pbeg !='\0') {
						pbeg++;
					}

					//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

					{  register unsigned long len;
					   unsigned char *w;

	 				   len = pbeg-codeStart;

					   ////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
					   writeWord(outWriter,codeStart, len);

					   getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);

						w=words[j].word;
						register int i;
						for (i=0;i<len;i++) {  //copia tambien el '\0'
							*(w++)=*(codeStart++);
						}
						*w='\0';
						words[j].weight = 0;
						words[j].size=len;
					}

					pbeg++;
					zeroNode++;
	             // fprintf(stderr,"\n Chega ao final");

				}/*else == zeronode*/



		/**/	{    // HACE EL UPDATE
				register unsigned long f,p,i,h;
				register unsigned long temp;
				unsigned char *tmpW;

					p=j;
					f = words[p].weight;
					words[p].weight +=1;

					i=j;
					j = topBlock[f];

					temp = words[i].weight;
					words[i].weight = words[j].weight;
					words[j].weight = temp;

					tmpW = words[i].word;
					words[i].word = words[j].word;
					words[j].word = tmpW;

					temp = words[i].size;
					words[i].size = words[j].size;
					words[j].size = temp;

					topBlock[f]=j+1;

					if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
						topBlock[f+1] = 0;
						MAXFREQ++;
					}
					j=i;
				}
		/**/

			T[0] +=oBytes(S_value -1, j);
			//T[1] +=oBytes(S_value, j);
			T[1] +=K;

			if (T[0] < T[1]){
				S_value --;
			  	C_value ++;
			  	T[1] = T[0];
//				printf ("*m2*CAMBIO DE S: VALOR BAJA DE [%lu] A [%lu]\n",S_value+1,S_value);
//				printf ("*m2*zeroNode vale [%lu]",zeroNode);
			}


		}/* while interno*/



   	T[2] = T[1];

/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/
/* TERCER MAIN */
/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/
/*===========================================*/



	//fprintf(stderr, "\n\n\nEntering General Main...");

	I[0][1]=W[S_value-1][0];
	I[1][0]=W[S_value+1][0];

	I[1][1]=W[S_value+1][1];
	I[2][0]=W[S_value-1][1];

	I[2][1]=W[S_value+1][2];
	I[3][0]=W[S_value-1][2];

	I[3][1]=W[S_value+1][3];

  	T[2] = T[1] =T[0] = 0;

/************************************************************************/
/************************************************************************/
/************************************************************************/

			while ( pbeg < pend )   {
				codeStart = pbeg;

				K=1;
				if ( (*pbeg) >= S_value ) {
					j = (*pbeg) - S_value;
					pbeg++;
					K++;

					while ((*pbeg) >= S_value) {

						j = j * C_value + ((*pbeg) - S_value);
						pbeg++;
						K++;
					}

					j = j * S_value + (*pbeg);
					if (K!=0)
					j += W[S_value][K-2];

				}
				else {
					j = (*pbeg);
				}


//				j=0;
//				while (*pbeg >=128) {
//					j = j * 128 + (*pbeg - 128);
//					pbeg++;
//				}
//				j = j * 128 + *pbeg;
//				j += TABLABASE[pbeg - codeStart];





				//fprintf(stderr,"\n decodificado=[%ld], bytes [%ld], con zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

				pbeg++;


	         if (j<zeroNode) {
					//	writes a word
					writeWord(outWriter,words[j].word, words[j].size);
				}
			 else { // j== zeronode

					codeStart = pbeg;
					while (*pbeg !='\0') {
						pbeg++;
					}

					//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

					{  register unsigned long len;
					   unsigned char *w;

	 				   len = pbeg-codeStart;

					   ////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
					   writeWord(outWriter,codeStart, len);

					   getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);

						w=words[j].word;
						register int i;
						for (i=0;i<len;i++) {  //copia tambien el '\0'
							*(w++)=*(codeStart++);
						}
						*w='\0';
						words[j].weight = 0;
						words[j].size=len;
					}

					pbeg++;
					zeroNode++;
	             // fprintf(stderr,"\n Chega ao final");

				}/*else == zeronode*/



		/**/	{    // HACE EL UPDATE
				register unsigned long f,p,i,h;
				register unsigned long temp;
				unsigned char *tmpW;

					p=j;
					f = words[p].weight;
					words[p].weight +=1;

					i=j;
					j = topBlock[f];

					temp = words[i].weight;
					words[i].weight = words[j].weight;
					words[j].weight = temp;

					tmpW = words[i].word;
					words[i].word = words[j].word;
					words[j].word = tmpW;

					temp = words[i].size;
					words[i].size = words[j].size;
					words[j].size = temp;

					topBlock[f]=j+1;

					if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
						topBlock[f+1] = 0;
						MAXFREQ++;
					}
					j=i;
				}
		/**/

				takeAccount();

			}/* while interno*/


		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);






/************************************************************************/
/************************************************************************/
/************************************************************************/
	while (!bufEmpty(B) ) {
	//fprintf(stderr, "\n paso 1 ");

		//fprintf(stderr, "\nEntra en while principal");

		bufCurrent (B,(char**)&text,(char**)&top);
		pbeg = text; pend = top;
	    //fprintf(stderr,"\n top = [%ld]",((unsigned long) top) - ((unsigned long)B->data));


		while ( pbeg < pend )   {
				codeStart = pbeg;

				K=1;
				if ( (*pbeg) >= S_value ) {
					j = (*pbeg) - S_value;
					pbeg++;
					K++;

					while ((*pbeg) >= S_value) {

						j = j * C_value + ((*pbeg) - S_value);
						pbeg++;
						K++;
					}

					j = j * S_value + (*pbeg);
					if (K!=0)
					j += W[S_value][K-2];

				}
				else {
					j = (*pbeg);
				}


//				j=0;
//				while (*pbeg >=128) {
//					j = j * 128 + (*pbeg - 128);
//					pbeg++;
//				}
//				j = j * 128 + *pbeg;
//				j += TABLABASE[pbeg - codeStart];





				//fprintf(stderr,"\n decodificado=[%ld], bytes [%ld], con zeronodoe = [%ld]",j,pbeg - codeStart +1,zeroNode);

				pbeg++;


	         if (j<zeroNode) {
					//	writes a word
					writeWord(outWriter,words[j].word, words[j].size);
				}
			 else { // j== zeronode

					codeStart = pbeg;
					while (*pbeg !='\0') {
						pbeg++;
					}

					//	insertElement ((unsigned char *) codeStart, pbeg-codeStart, j);

					{  register unsigned long len;
					   unsigned char *w;

	 				   len = pbeg-codeStart;

					   ////fprintf(stderr,"\nzeroNode= true wordLeng=[%ld]",pbeg-codeStart);
					   writeWord(outWriter,codeStart, len);

					   getMemoryBlock(_memMgr,( char**)&(words[j].word),len+1);

						w=words[j].word;
						register int i;
						for (i=0;i<len;i++) {  //copia tambien el '\0'
							*(w++)=*(codeStart++);
						}
						*w='\0';
						words[j].weight = 0;
						words[j].size=len;
					}

					pbeg++;
					zeroNode++;
	             // fprintf(stderr,"\n Chega ao final");

				}/*else == zeronode*/



		/**/	{    // HACE EL UPDATE
				register unsigned long f,p,i,h;
				register unsigned long temp;
				unsigned char *tmpW;

					p=j;
					f = words[p].weight;
					words[p].weight +=1;

					i=j;
					j = topBlock[f];

					temp = words[i].weight;
					words[i].weight = words[j].weight;
					words[j].weight = temp;

					tmpW = words[i].word;
					words[i].word = words[j].word;
					words[j].word = tmpW;

					temp = words[i].size;
					words[i].size = words[j].size;
					words[j].size = temp;

					topBlock[f]=j+1;

					if (MAXFREQ == f+1) {  //The most frequent words need to be increased.
						topBlock[f+1] = 0;
						MAXFREQ++;
					}
					j=i;
				}
		/**/

				takeAccount();


			}/* while interno*/


		nextbuf=pbeg;
		bufLoad (B, (char*)nextbuf);


	}//while



	//fprintf(stderr, "\n... DECOMPRESSION FINISHED after %d words\n\n");

	bufDestroy(B);
	destroyWriter(outWriter);
	free(words);


} //mainDecompressor



/*------------------------------------------------------------------ */
/*
 *  DECOMPRESSOR PROGRAM OF DYNAMIC (S,C)-DC.
 */
/*------------------------------------------------------------------ */
int main(int argc, char* argv[])
{
	char fileName[20];

	long sizeNValue;
	long sizeNValue2;
	int needToLoadSizes=0;
	

	showMessage();

	if (argc < 3) {
		fprintf(stderr,"\n Wrong call to DSCDC dynamic decompressor\n");
		fprintf(stderr," Use:  %s  <in file>  <out file> [<VOC SIZE]>[<TOP SIZE]>\n",argv[0]);
		return 0;
	}


	sizeNValue = MAX_LEN_VOCABULARY;
	sizeNValue2 = INITIAL_TOP_FREQ;
	
    if ((argc ==3) || (argc = 4)) 
    	needToLoadSizes = 1;


	//	if ((argc ==3) || (argc = 4))
	//		readSizeVocabularyandTop(argv[1],&sizeNValue,&sizeNValue2);
		
	//	if (argc ==4)
	//		sizeNValue = atol(argv[3]);
	
	if (argc ==5) {
		sizeNValue = atol(argv[3]);
		sizeNValue2 = atol(argv[4]);
		needToLoadSizes = 0;
	}

	//printf("\n VOC SIZE [%ld],     TOP SIZE [%ld] \n",sizeNValue,sizeNValue2);

	StartRusageTime();
	mainDecompressor(argv[1], argv[2],needToLoadSizes,sizeNValue,sizeNValue2);
	StopRusageTimeFile("timesDecDSCDC.txt",argv[1]);
	fprintf(stderr,"\n ZERONODE VALUE AT THE END IS:   %ld \n",zeroNode);
}
